/*****************************************************************************/
/*                                                                           */
/* Author: Michael C. Nechyba                                                */
/* Last modified: 02/16/2004                                                 */
/*                                                                           */
/* File: "dwt_util.c"                                                        */
/*                                                                           */
/*****************************************************************************/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include "file_io.h"
#include "matrix.h"
#include "gen_math.h"
#include "dwt_util.h"

/****************************** Subroutines **********************************/

/*****************************************************************************/

int haar_filter(double *lpf)
{
  lpf[0] = 1./sqrt(2.);
  lpf[1] = lpf[0];

  return(2);
}

/*****************************************************************************/

int daubechies_filter(double *lpf, int n)
{
  switch(n) {
  case 2: {
    haar_filter(lpf);
    break;
  }
  case 4: {
    lpf[0] = 0.482962913144534143375;
    lpf[1] = 0.83651630373780790558;
    lpf[2] = 0.22414386804201338103;
    lpf[3] = -0.12940952255126038117;
    break;
  }
  case 6: {
    lpf[0] = 0.33267055295008261600;
    lpf[1] = 0.80689150931109257649;
    lpf[2] = 0.45987750211849157010;
    lpf[3] = -0.13501102001025458870;
    lpf[4] = -0.08544127388202666169;
    lpf[5] = 0.035226291885709536603;
    break;
  }
  case 8: {
    lpf[0] = 0.23037781330889650086;
    lpf[1] = 0.71484657055291564709;
    lpf[2] = 0.6308807679298589079;
    lpf[3] = -0.0279837694168598542;
    lpf[4] = -0.1870348117190930841;
    lpf[5] = 0.0308413818355607636;
    lpf[6] = 0.03288301166688519974;
    lpf[7] = -0.010597401785069032105;
    break;
  }
  case 10: {
    lpf[0] = 0.16010239797419291448;
    lpf[1] = 0.6038292697971896705;
    lpf[2] = 0.7243085284377729277;
    lpf[3] = 0.1384281459013207315;
    lpf[4] = -0.2422948870663820319;
    lpf[5] = -0.0322448695846383746;
    lpf[6] = 0.0775714938400457135;
    lpf[7] = -0.0062414902127982743;
    lpf[8] = -0.01258075199908199947;
    lpf[9] = 0.003335725285473771278;
    break;
  } 
  case 12: {
    lpf[0] = 0.11154074335010946362;
    lpf[1] = 0.4946238903984530857;
    lpf[2] = 0.7511339080210953507;
    lpf[3] = 0.3152503517091976291;
    lpf[4] = -0.226264693965439820;
    lpf[5] = -0.129766867567261936;
    lpf[6] = 0.097501605587323049;
    lpf[7] = 0.0275228655303057286;
    lpf[8] = -0.0315820393174860296;
    lpf[9] = 0.00055384220116149614;
    lpf[10] = 0.00477725751094551064;
    lpf[11] = -0.001077301085308479565;
    break;
  }
  default: {
    fprintf(stderr, "Daubechies filter size should be 4, 6, 8, 10 or 12.\n");
    exit(1);
    break;
  }
  }
  
  return(n);
}

/*****************************************************************************/

int highpass_filter(double *hpf, double *lpf, int n)
{
  int i;
  double sign = -1.;

  for(i = 0; i < n; i++) {
    hpf[n-i-1] = sign*lpf[i];
    sign *= -1.;
  }

  return(1);
}

/*****************************************************************************/

int inverse_filter(double *ilpf, double *lpf, int n)
{
  int i;

  for(i = 0; i < n; i++) {
    if(i % 2 == 1)
      ilpf[i] = lpf[i];
    else
      ilpf[i] = lpf[n-i-2];
  }

  return(1);
}

/*****************************************************************************/

int setup_forward_filters(double *lpf, double *hpf, int nf)
{
  daubechies_filter(lpf, nf);
  highpass_filter(hpf, lpf, nf);

  return(1);
}

/*****************************************************************************/

int setup_inverse_filters(double *ilpf, double *ihpf, int nf)
{
  double lpf[MAX_DWT_FILTER_LENGTH];

  daubechies_filter(&lpf[0], nf);
  inverse_filter(ilpf, &lpf[0], nf);
  highpass_filter(ihpf, ilpf, nf);

  return(1);
}

/*****************************************************************************/

int dwt1D_level(double *w, double *x, int nx, 
		double *lpf, double *hpf, int nf,
		int level)
{
  int i, j, n, n2, offset;
  
  n = nx;
  for(i = 0; i < level-1; i++)
    n /= 2;

  n2 = n/2;

  for(i = 0; i < n2; i++) {
    w[i] = 0.; 
    w[i + n2] = 0.;
    
    for(j = 0; j < nf; j++) {
      offset = (2*i + j) % n;
      w[i] += lpf[j]*x[offset];
      w[i + n2] += hpf[j]*x[offset];
    }
  }

  return(1);
}

/*****************************************************************************/

int dwt1D(double *x, int nx,
	  double *lpf, double *hpf, int nf,
	  int level)
{
  int k;
  double *w;

  w = allocate_vector(w, nx);

  for(k = 0; k < level; k++) {
    dwt1D_level(w, x, nx, lpf, hpf, nf, k+1);
    copy_vector(w, x, nx);
  }

  free_vector(w);
  return(1);
}

/*****************************************************************************/

int idwt1D_level(double *x, double *w, int nw, 
		 double *lpf, double *hpf, int nf,
		 int level)
{
  int i, j, n, n2, offset, *indx;

  n = nw;
  for(i = 0; i < level-1; i++)
    n /= 2;

  n2 = n/2;

  indx = int_allocate_vector(indx, n);
  for(i = 0; i < n2; i++) {
    indx[2*i] = i;
    indx[2*i+1] = i + n2;
  }

  for(i = 0; i < n; i+=2) {
    x[i] = 0.; 
    x[i + 1] = 0.;
    
    for(j = 0; j < nf; j++) {
      offset = indx[(n-(nf-2)+i+j)%n];
      x[i] += lpf[j]*w[offset];
      x[i + 1] += hpf[j]*w[offset];
    }
  }

  for(i = n; i < nw; i++)
    x[i] = w[i];

  free(indx);
  return(1);
}

/*****************************************************************************/

int idwt1D(double *w, int nw,
	   double *lpf, double *hpf, int nf,
	   int level)
{
  int k;
  double *x;

  x = allocate_vector(x, nw);

  for(k = level; k > 0; k--) {
    idwt1D_level(x, w, nw, lpf, hpf, nf, k);
    copy_vector(x, w, nw);
  }

  free_vector(x);
  return(1);
}

/*****************************************************************************/

int dwt2D(double **x, int nx, double *lpf, int nf, int level)
{
  int i, k, n;
  double *hpf;
  double **w;
  
  n = nx;

  w = allocate_matrix(w, n, n);
  hpf = allocate_vector(hpf, nf);

  highpass_filter(hpf, lpf, nf);

  for(k = 0; k < level; k++) {
    for(i = 0; i < n; i++)
      dwt1D(&x[i][0], n, lpf, hpf, nf, 1);
    
    transpose_matrix(x, n, n, w);
    
    for(i = 0; i < n; i++)
      dwt1D(&w[i][0], n, lpf, hpf, nf, 1);
    
    transpose_matrix(w, n, n, x);

    n /= 2;
  }
    
  free_vector(hpf);
  free_matrix(w, nx);

  return(1);
}

/*****************************************************************************/

int idwt2D(double **w, int nw, double *lpf, int nf, int level)
{
  int i, k, n;
  double *ilpf, *ihpf;
  double **x;
  
  x = allocate_matrix(x, nw, nw);
  ilpf = allocate_vector(ilpf, nf);
  ihpf = allocate_vector(ihpf, nf);

  inverse_filter(ilpf, lpf, nf);
  highpass_filter(ihpf, ilpf, nf);

  n = nw;
  for(k = 1; k < level; k++)
    n /= 2;

  for(k = 0; k < level; k++) {
    transpose_matrix(w, nw, nw, x);

    for(i = 0; i < n; i++)
      idwt1D(&x[i][0], n, ilpf, ihpf, nf, 1);
    
    transpose_matrix(x, nw, nw, w);
    
    for(i = 0; i < n; i++)
      idwt1D(&w[i][0], n, ilpf, ihpf, nf, 1);

    n *= 2;
  }
    
  free_vector(ihpf);
  free_vector(ilpf);
  free_matrix(x, nw);

  return(1);
}

/*****************************************************************************/

int max_levels(int n, int nf)
{
  int count, level = 1;

  count = n;
  while(count > nf) {
    count /= 2;
    level++;
  }

  return(level);
}

/*****************************************************************************/

int zero_coefficients2D(double **w, int n, double thresh)
{
  int i, j, n_coeff = 0;

  for(i = 0; i < n; i++) {
    for(j = 0; j < n; j++) {
      if(fabs(w[i][j]) < thresh) {
	w[i][j] = 0.;
	n_coeff++;
      }
    }
  }

  return(n_coeff);
}

/*****************************************************************************/

int extract_band(double **b, int wb, int wlevel, double **w, int nw)
{
  int i, n, n2, xs, ys, xe, ye;

  n = nw;
  for(i = 1; i < wlevel; i++)
    n /= 2;

  n2 = n/2;

  switch(wb) {
  case LL_BAND: {
    xs = 0;  ys = 0;
    xe = n2; ye = n2;
    break;
  }
  case LH_BAND: {
    xs = n2; ys = 0;
    xe = n;  ye = n2;
    break;
  }
  case HL_BAND: {
    xs = 0;  ys = n2;
    xe = n2; ye = n;
    break;
  }
  case HH_BAND: {
    xs = n2; ys = n2;
    xe = n;  ye = n;
    break;
  }
  }

  extract_submatrix(b, w, xs, ys, xe, ye);
  return(n2);
}

/*****************************************************************************/

int extract_submatrix(double **sm, double **m,
		      int xs, int ys, int xe, int ye)
{
  int i, j;

  for(i = ys; i < ye; i++)
    for(j = xs; j < xe; j++)
      sm[i-ys][j-xs] = m[i][j];

  return(1);
}

/*****************************************************************************/
