/*****************************************************************************/
/*                                                                           */
/* Title: feature extraction from n x n image (PPM/P6 RGB format)            */
/*                                                                           */
/* File: "featuresDWT.c"                                                     */
/*                                                                           */
/* Author: Michael C. Nechyba                                                */
/* Last modified: 02/16/2004                                                 */
/*                                                                           */
/*****************************************************************************/

/*
  This program takes as input an n x n input image I (PPM/P6/RGB
  format), converts it to an intensity image, subdivides it into
  windows W of m x m, m << n, with overlap ov, computes a k-level 2D
  DWT for each window W, and then saves all the separate bands to
  files with the same name as the input image, except with the level
  and band replacing the .ppm extension. The -noabs flag controls
  whether bands are saved as signed values or abs values.

  This does exactly the same thing as the Mathematica function
  GenerateSixBandTrainingData[] for the texture segmentation example
  for k = 2, m = 16 and ov = 4, except it is much faster.
*/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include "arg.h"
#include "matrix.h"
#include "file_io.h"
#include "gen_math.h"
#include "dwt_util.h"
#include "ppm_util.h"


int extract_features(char *output_file, PPM *image,
		     int level, int nf, int nwin, int overlap, int noabs,
		     int precision);
int display_matrix_as_vector(FILE *fp, double **matrix, 
			     int n, int m, int noabs, int precision);

PPM Image;

/**************************** Main Program ***********************************/

int main(int argc, char **argv)
{
  char *image_file = NULL, output_file[256];
  int filename_length;
  int nf = 4, level = 2;
  int nwin = 16, overlap = 4;
  int noabs = 0;
  int precision = 15;

  arg_parse(argc, argv,
            "", "Usage:",
	    "-image %S", &image_file, "set image file",
	    "-level [%d]", &level, "set number of levels [%d]", level,
    	    "-window [%d]", &nwin, "set window size [%d]", nwin,
	    "-overlap [%d]", &overlap, "set overlap between windows [%d]",
	    overlap,
	    "-filter [%d]", &nf, 
	    "set filter length [%d] (2: Haar, 4,6,8,10,12: Daubechies-n)", nf,
            "-noabs", ARG_FLAG(&noabs), 
	    "don't take the absolute value of wavelet coefficients",
	    "-precision [%d]", &precision, "set output precision [%d]",
	    precision,
            NULL);
  
  if(nf % 2 != 0 || nf < 2 || nf > 12) {
    fprintf(stderr, 
	    "Filter length has to be 2, 4, 6, 8, 10 or 12. Exiting.\n");
    exit(-1);
  }
  
  if(image_file == NULL) {
    fprintf(stderr, "No image file specified. Exiting.\n");
    exit(-1);
  }

  /* read PPM image file */

  read_ppm_file(image_file, &Image);

  filename_length = strlen(image_file)-4;
  strncpy(&output_file[0], image_file, filename_length);
  output_file[filename_length] = (char) 0;
  
  extract_features(&output_file[0], &Image, level, nf, nwin, overlap, noabs,
		   precision);

  exit(1);
}

/*****************************************************************************/

#define MAX_BANDS 20

int extract_features(char *output_file, PPM *image,
		     int level, int nf, int nwin, int overlap, int noabs,
		     int precision)
{
  int n, m, i, j, k, ns, ms, nb, wb, wl, nband;
  double **x, **window, **band, *lpf;
  FILE *fp[MAX_BANDS];
  char str[256];

  nb = level*3 + 1;
  for(k = 0; k < nb; k++) {
    sprintf(&str[0], "%s.%02d", output_file, k+1);
    fp[k] = fopen(&str[0], "w");
  }

  lpf = allocate_vector(lpf, nf);
  window = allocate_matrix(window, nwin, nwin);
  band = allocate_matrix(band, nwin, nwin);

  n = image->ydim;
  m = image->xdim;

  x = convert_ppm_to_double(x, image, 0);
  daubechies_filter(lpf, nf);

  ns = n - nwin + 1;
  ms = m - nwin + 1;

  for(i = 0; i < ns; i += overlap) {
    for(j = 0; j < ms; j += overlap) {
      extract_submatrix(window, x, j, i, j+nwin, i+nwin);
      dwt2D(window, nwin, lpf, nf, level);
      for(k = 0; k < nb; k++) {
	if(k == nb - 1) {
	  wl = level;
	  wb = LL_BAND;
	} else {
	  wl = (k/3) + 1;
	  wb = (k%3) + 1;
	}
	nband = extract_band(band, wb, wl, window, nwin);
	display_matrix_as_vector(fp[k], band, nband, nband, noabs, precision);
      }
    }
  }

  for(k = 0; k < nb; k++) {
    fclose(fp[k]);
  }

  free_matrix(x, n);
  free_matrix(band, nwin);
  free_matrix(window, nwin);
  free_vector(lpf);

  return(1);
}

/*****************************************************************************/

int display_matrix_as_vector(FILE *fp, double **matrix, 
			     int n, int m, int noabs, int precision)
{
  char format[STR_REG], format2[STR_REG];
  int i, j;

  if(precision <= 0)
    precision = 10;

  sprintf(format, "%%.%df ", precision);

  for(i = 0; i < n; i++) {
    for(j = 0; j < m; j++) {
      if(noabs) {
	fprintf(fp, format, matrix[i][j]);
      } else {
	fprintf(fp, format, fabs(matrix[i][j]));
      }
    }
  }

  fprintf(fp, "\n");
  return(1);
}

/*****************************************************************************/
