#!/usr/bin/perl
###########################################################################
#
# Usage: ../../src/texture.pl parameter_file (in demos/set[12] directories)
#
# Description: 
#              
# This perl script performs texture segmentation. Through a parameter
# file, it takes as input a set of training images (PPM) and test
# images (PPM) of different textures, performs feature extraction on
# the training images using a multilevel DWT (as described in class
# and the notes), computes VQ codebooks over the various DWT bands,
# and generates the VQ labeling of the training data based on those VQ
# codebooks. It also generates the VQ labeling of the test data.  #
#
# The ouput of this script is a set of VQ codebooks (one for each
# feature band), and a set of label files for each image and each
# feature band. The Mathematica notebook 'visualize_results.nb' takes
# these files, computes histograms and performs final texture
# segmentation on the test images.
#
# System calls: /bin/rm, /bin/cat
#
# C calls: ../../bin/featuresDWT, ../../bin/codebook, ../../bin/vq
#
###########################################################################
#
# Parameters: All possible parameters are given below, with defaults
# in square brackets, if applicable.
#
# TrainingFiles <ppm image filename list>
# TestFiles <ppm image filename list>
# DWTLevels <levels [2]>
# WindowSize <feature window size [16]>
# WindowOffset <offset between consecutive feature windows [4]>
# FilterSize <filter size [4]>
# AbsoluteValue <absolute value or signed coefficients [1]>
# KeepLLBand <keep LL band as a feature [0]>
# VQLevels <log2 number of VQ levels [5]>
# VQConvergence <convergence criterion for VQ algorithm [0.00001]>
# Precision <output precision [15]>
#
###########################################################################

$printOnly = 0;

if(!$printOnly) {
    system("echo `date`");
}

###########################################################################
#
# Variable initialization and default settings
#
###########################################################################

$parameter = $ARGV[0];

$bindir = "../../bin";
$precision = 15;

$haveTrainingFiles = "none";
$haveTestFiles = "none";

$levels = 2;
$windowSize = 16;
$filterSize = 4;
$absoluteValue = 1;
$keepLLBand = 0;
$vqLevels = 5;
$vqConvergence = 0.00001;

###########################################################################
#
# Read in parameter file
#
###########################################################################

open parameter;

while(<parameter>) {	
    chop;		
    @a = split;
    
    if ($a[0] =~ "#") {
	# do nothing
    } elsif ($a[0] =~ "TrainingFiles") {
	@trainingFiles = @a[1..$#a];
	$haveTrainingFiles = "ok";
    } elsif ($a[0] =~ "TestFiles") {
	@testFiles = @a[1..$#a];
	$haveTestFiles = "ok";
    } elsif ($a[0] =~ "DWTLevels") {
	$levels = $a[1];
    } elsif ($a[0] =~ "WindowSize") {
	$windowSize = $a[1];
    } elsif ($a[0] =~ "WindowOffset") {
	$windowOffset = $a[1];
    } elsif ($a[0] =~ "FilterSize") {
	$filterSize = $a[1];
    } elsif ($a[0] =~ "AbsoluteValue") {
	$absoluteValue = $a[1];
	if($absoluteValue == 0) {
	    $noabs = " -noabs";
	} else {
	    $noabs = " ";
	}
    } elsif ($a[0] =~ "KeepLLBand") {
	$keepLLBand = $a[1];
    } elsif ($a[0] =~ "VQLevels") {
	$vqLevels = $a[1];
    } elsif ($a[0] =~ "VQConvergence") {
	$vqConvergence = $a[1];
    } elsif ($a[0] =~ "Precision") {
	$precision = $a[1];
    } else {
	print "Warning: did not recognize $a[0].\n";
    }
}

close parameter;

if($haveTrainingFiles =~ "none" || $haveTestFiles =~ "none") {
    print "No training and/or test files given. Exiting.\n";
    exit;
}

###########################################################################
#
# Further variable initialization
#
###########################################################################

$rm = "/bin/rm";
$cat = "/bin/cat";

$featuresDWT = join('', $bindir, "/featuresDWT");
$codebook = join('', $bindir, "/codebook");
$vq = join('', $bindir, "/vq");

###########################################################################
# TRAINING
###########################################################################
#
# Extract feature vectors from training images
#
###########################################################################

for($i = 0; $i <= $#trainingFiles; $i++) {
    $sys = join('', $featuresDWT,
		" -image $trainingFiles[$i]",
		" -level $levels",
		" -window $windowSize",
		" -overlap $windowOffset",
		" -filter $filterSize",
		$noabs,
		" -precision $precision");

    if(!$printOnly) {
	system($sys);
    } else {
	print "$sys\n";
    }
}

###########################################################################
#
# Compute # of generated bands and toss out highest-level LL if desired
#
###########################################################################

$nBands = $levels*3 + 1;

if(!$keepLLBand) {
    $sys = sprintf "%s *.%02d", $rm, $nBands;
    $nBands = $nBands - 1;
   
    if(!$printOnly) {
	system($sys);
    } else {
	print "$sys\n";
    }
}

###########################################################################
#
# Concatenate band files across training images
#
###########################################################################

for($i = 1; $i <= $nBands; $i++) {
    $sys = sprintf "%s *.%02d > band.%02d.vq", $cat, $i, $i;

    if(!$printOnly) {
	system($sys);
    } else {
	print "$sys\n";
    }
}

###########################################################################
#
# Generate VQ codebooks
#
###########################################################################

for($i = 1; $i <= $nBands; $i++) {
    $file = sprintf "band.%02d.vq", $i;
    open file;
    $line = <file>;
    chop($line); chop($line);
    @tmp = split(/ /, $line);
    close file;

    $vqLength = $#tmp + 1;
    $outputFile = sprintf "band.%02d.codes", $i;

    $sys = join('', $codebook,
		" -input $file",
		" -output $outputFile",
		" -convergence $vqConvergence",
		" -length $vqLength",
		" -nCodes $vqLevels",
		" -precision $precision");

    if(!$printOnly) {
	system($sys);
    } else {
	print "$sys\n";
    }
}

###########################################################################
#
# Remove VQ training data
#
###########################################################################

$sys = sprintf "%s *.vq", $rm;

if(!$printOnly) {
    system($sys);
} else {
#    print "$sys\n";
}

###########################################################################
#
# Generate VQ labels for training data
#
###########################################################################

for($k = 0; $k <= $#trainingFiles; $k++) {
    @tmp = split(/\./,$trainingFiles[$k]);
    $imagePrefix = $tmp[0];
    for($i = 1; $i <= $nBands; $i++) {
	$file = sprintf "%s.%02d", $imagePrefix, $i;
	open file;
	$line = <file>;
	chop($line); chop($line);
	@tmp = split(/ /, $line);
	close file;

	$vqLength = $#tmp + 1;
	$outputFile = sprintf "%s.%02d.labels", $imagePrefix, $i;
	$codesFile = sprintf "band.%02d.codes.%d", $i, $vqLevels;

	$sys = join('', $vq,
		    " -i $file",
		    " -o $outputFile",
		    " -c $codesFile",
		    " -l $vqLength");
	
	if(!$printOnly) {
	    system($sys);
	} else {
	    print "$sys\n";
	}

	$sys = sprintf "%s %s", $rm, $file;
	
	if(!$printOnly) {
	    system($sys);
	} else {
	    print "$sys\n";
	}
    }
}

###########################################################################
#
# Clean up intermediate VQ codebook files 
#
###########################################################################

$sys = sprintf "%s *.codes.[%d-%d]", $rm, 1, $vqLevels-1;

if(!$printOnly) {
    system($sys);
} else {
    print "$sys\n";
}

###########################################################################
# TESTING
###########################################################################
#
# Extract feature vectors from test images
#
###########################################################################

for($i = 0; $i <= $#testFiles; $i++) {
    $sys = join('', $featuresDWT,
		" -image $testFiles[$i]",
		" -level $levels",
		" -window $windowSize",
		" -overlap $windowOffset",
		" -filter $filterSize",
		$noabs,
		" -precision $precision");

    if(!$printOnly) {
	system($sys);
    } else {
	print "$sys\n";
    }
}

###########################################################################
#
# Compute # of generated bands and toss out highest-level LL if desired
#
###########################################################################

$nBands = $levels*3 + 1;

if(!$keepLLBand) {
    $sys = sprintf "%s *.%02d", $rm, $nBands;
    $nBands = $nBands - 1;
   
    if(!$printOnly) {
	system($sys);
    } else {
	print "$sys\n";
    }
}

###########################################################################
#
# Generate VQ labels for testing data
#
###########################################################################

for($k = 0; $k <= $#testFiles; $k++) {
    @tmp = split(/\./,$testFiles[$k]);
    $imagePrefix = $tmp[0];
    for($i = 1; $i <= $nBands; $i++) {
	$file = sprintf "%s.%02d", $imagePrefix, $i;
	open file;
	$line = <file>;
	chop($line); chop($line);
	@tmp = split(/ /, $line);
	close file;

	$vqLength = $#tmp + 1;
	$outputFile = sprintf "%s.%02d.labels", $imagePrefix, $i;
	$codesFile = sprintf "band.%02d.codes.%d", $i, $vqLevels;

	$sys = join('', $vq,
		    " -i $file",
		    " -o $outputFile",
		    " -c $codesFile",
		    " -l $vqLength");
	
	if(!$printOnly) {
	    system($sys);
	} else {
	    print "$sys\n";
	}

	$sys = sprintf "%s %s", $rm, $file;
	
	if(!$printOnly) {
	    system($sys);
	} else {
	    print "$sys\n";
	}
    }
}

if(!$printOnly) {
    system("echo `date`");
}

exit;
